﻿using System;
using System.Reflection;

namespace Sheepy.PhoenixPt_SpillExpGlobal {
   using Harmony;
   using PhoenixPoint.Common.Entities;
   using PhoenixPoint.Common.Entities.Characters;
   using PhoenixPoint.Common.Levels.Missions;
   using PhoenixPoint.Geoscape.Entities;
   using PhoenixPoint.Tactical.Entities;
   using PhoenixPoint.Tactical.Levels;
   using PhoenixPoint.Tactical.View.ViewControllers;
   using PhoenixPoint.Tactical.View.ViewModules;
   using System.Collections.Generic;
   using static System.Reflection.BindingFlags;

   public class Mod {
      //private static Logger Log;

      public static void Init () {
         //Log = new Logger( "Mods/SheepyMods.log" ){ Prefix = typeof( Mod ).Namespace + " " };
         HarmonyInstance harmony = HarmonyInstance.Create( typeof( Mod ).Namespace );

         Patch( harmony, typeof( TacticalFaction ), "GiveExperienceForObjectives", "Prefix_GiveExp", "Postfix_GiveExp" );
         Patch( harmony, typeof( LevelProgression ), "AddExperience", null, "Postfix_AddExp" );
         Patch( harmony, typeof( UIModuleBattleSummary ), "SetProgressSlider", null, "Postfix_UISoldier" );
         Patch( harmony, typeof( GeoCharacter ), "ApplyTacInstanceData", null, "Postfix_ApplyExp" );
      }

      #region Harmony helpers
      private static void Patch ( HarmonyInstance harmony, Type target, string toPatch, string prefix, string postfix = null ) {
         Patch( harmony, target.GetMethod( toPatch, Public | NonPublic | Instance | Static ), prefix, postfix );
      }

      private static void Patch ( HarmonyInstance harmony, MethodInfo toPatch, string prefix, string postfix = null ) {
         harmony.Patch( toPatch, ToHarmonyMethod( prefix ), ToHarmonyMethod( postfix ) );
      }

      private static HarmonyMethod ToHarmonyMethod ( string name ) {
         if ( name == null ) return null;
         MethodInfo func = typeof( Mod ).GetMethod( name );
         if ( func == null ) throw new NullReferenceException( name + " is null" );
         return new HarmonyMethod( func );
      }
      #endregion

      private static bool captureExp = false;
      private static Dictionary<LevelProgression, int> missionExp;
      private static Dictionary<GeoTacUnitId, int> spillSp;

      [ HarmonyPriority( Priority.VeryHigh ) ]
      public static void Prefix_GiveExp ( TacticalFaction __instance ) { try {
         if ( __instance?.ParticipantKind != TacMissionParticipant.Player ) return;
         missionExp = new Dictionary<LevelProgression, int>( 16 );
         captureExp = true;
      // } catch ( Exception ex ) { Log.Error( ex ); } }
      } catch ( Exception ) { } }

      // Capture exp awarded for each soldier
      public static void Postfix_AddExp ( LevelProgression __instance, int experience ) { try {
         if ( ! captureExp || missionExp == null ) return;
         missionExp.TryGetValue( __instance, out int exp );
         missionExp[ __instance ] = exp + experience;
      // } catch ( Exception ex ) { Log.Error( ex ); } }
      } catch ( Exception ) { } }

      // Check overflow and convert to SP
      public static void Postfix_GiveExp ( TacticalFaction __instance ) { try {
         if ( missionExp == null ) return;
         captureExp = false;
         float spConversionRate = __instance.TacticalLevel?.SharedData?.ExpConvertedToSkillpoints ?? 0.01f;
         spillSp = new Dictionary<GeoTacUnitId, int>( missionExp.Count );
         foreach ( TacticalActor actor in __instance.TacticalActors ) {
            LevelProgression lv = actor?.LevelProgression;
            if ( lv == null || ! actor.IsAlive || ! missionExp.TryGetValue( lv, out int exp ) ) continue;
            int earned = lv.Experience - lv.ExperienceReference;
            int spill = exp - earned;
            int sp = Math.Max( 0, (int) Math.Floor( spill * spConversionRate * 2 ) );
            if ( sp > 0 ) {
               spillSp[ actor.GeoUnitId ] = sp;
               //Log.Info( "(Tac) {0} Exp {1} = Earned {2} + Overflow {3}, spill into {4} sp", actor.GetDisplayName(), exp, earned, spill, sp );
            }
         }
         missionExp = null;
         if ( spillSp.Count <= 0 ) spillSp = null;
      // } catch ( Exception ex ) { Log.Error( ex ); } }
      } catch ( Exception ) { } }

      // Show SP gain on mission result screen
      public static void Postfix_UISoldier ( SoldierResultElement soldierResultElement ) { try {
         TacticalActor actor = soldierResultElement.Actor;
         LevelProgression lv = actor?.LevelProgression;
         if ( lv == null || ! actor.IsAlive || lv.Experience != lv.ExperienceReference ) return;

         if ( ! spillSp.TryGetValue( actor.GeoUnitId, out int sp ) ) return;
         soldierResultElement.EarnedExperience.text = String.Format( "SP +{0}", sp );
         soldierResultElement.EarnedExperience.color = soldierResultElement.DefaultStatus.TextColor;
      // } catch ( Exception ex ) { Log.Error( ex ); } }
      } catch ( Exception ) { } }

      // Award SP to soldiers when on the way back to geoscope
      public static void Postfix_ApplyExp ( GeoCharacter __instance, TacActorUnitResult result ) { try {
         if ( spillSp == null || ! spillSp.TryGetValue( result.GeoUnitId, out int sp ) ) return;
         if ( sp > 0 ) {
            __instance.Progression.SkillPoints += sp;
            //Log.Info( "(Geo) {0} Skill Point +{1}", __instance.GetName(), sp );
            spillSp.Remove( result.GeoUnitId );
            if ( spillSp.Count <= 0 ) spillSp = null;
         }
      // } catch ( Exception ex ) { Log.Error( ex ); } }
      } catch ( Exception ) { } }
   }
}